package com.kabronkline.jft.service;

import java.util.ArrayList;
import java.util.List;
import java.util.Queue;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.PriorityBlockingQueue;

import org.apache.commons.logging.LogFactory;
import org.apache.commons.vfs2.FileObject;
import org.apache.commons.vfs2.FileSystemException;
import org.apache.commons.vfs2.FileSystemManager;
import org.apache.commons.vfs2.FileSystemOptions;
import org.apache.commons.vfs2.VFS;
import org.apache.commons.vfs2.impl.StandardFileSystemManager;
import org.apache.log4j.Logger;

import com.kabronkline.jft.exception.FileSystemServiceException;
import com.kabronkline.jft.transfer.FileTransfer;
import com.kabronkline.jft.transfer.FileTransferThread;

public abstract class FileSystemService {
	private static final Logger log = Logger.getLogger(FileSystemService.class);
	
	protected FileSystemOptions fileSystemOptions;
	protected int maxTransfers;
	protected ExecutorService transferService;
	protected Queue<FileTransfer> transferQueue;
	private final ThreadLocal<FileSystemManager> fileSystemManager = new ThreadLocal<FileSystemManager>() { 
	    @Override 
	    protected FileSystemManager initialValue() { 
	        StandardFileSystemManager mgr = new StandardFileSystemManager(); 
	        mgr.setLogger(LogFactory.getLog(StandardFileSystemManager.class)); 
	        try { 
	            mgr.init(); 
	        } 
	        catch (FileSystemException e) { 
	        	log.error("Unable to instantiate FileSystemManager", e);
	        } 
	        return mgr; 
	    } 
	};
	
	public ThreadLocal<FileSystemManager> getFileSystemManager() {
		return fileSystemManager;
	}

	public FileSystemService(int managerPoolSize) {
		this.transferQueue = new PriorityBlockingQueue<FileTransfer>();
		this.maxTransfers = managerPoolSize;
		this.transferService = Executors.newFixedThreadPool(managerPoolSize);
	}

	protected FileSystemOptions createFileSystemOptions() { 
		return new FileSystemOptions(); 
	}
	
	protected String getFilePathURI(String filePathURI) {
		return filePathURI;
	}
	
	protected FileObject retrieveFile(String filePathURI) throws FileSystemServiceException {
		if (fileSystemOptions == null) this.fileSystemOptions = createFileSystemOptions();
		FileSystemManager fileSystemManager = getFileSystemManager().get();
		FileObject fileObj = null;
		try {
			fileObj = fileSystemManager.resolveFile(getFilePathURI(filePathURI), fileSystemOptions);
		} catch (FileSystemException e) {
			log.error("Failed to retrieve file at URI : " + filePathURI, e);
			throw new FileSystemServiceException(e);
		}
		return fileObj;
	}
	
	protected void transferFile(FileObject sourceFile, FileObject destinationFile) 
	throws FileSystemServiceException {
		transferQueue.add(new FileTransfer(sourceFile, destinationFile));
		try {
			transferService.submit(new FileTransferThread(transferQueue)).get();
		} catch (InterruptedException e) {
			log.error("Failed to submit FileTransferThread to ExecutorService.", e);
			throw new FileSystemServiceException(e);
		} catch (ExecutionException e) {
			log.error("Failed to submit FileTransferThread to ExecutorService.", e);
			throw new FileSystemServiceException(e);
		}
	}
	
	protected void transferFiles(List<FileTransfer> fileTransfers) throws FileSystemServiceException {
		List<Future<List<FileTransfer>>> futureFileTransfers = new ArrayList<Future<List<FileTransfer>>>();
		for (FileTransfer fileTransfer : fileTransfers) {
			transferQueue.add(fileTransfer);
		}
		for (int i = 0; i < maxTransfers; i++) {
			futureFileTransfers.add(transferService.submit(new FileTransferThread(transferQueue)));
		}
		for (Future<List<FileTransfer>> futureTransfer : futureFileTransfers) {
			try {
				futureTransfer.get();
			} catch (InterruptedException e) {
				log.error("Problem encountered during file transfer.", e);
				throw new FileSystemServiceException(e);
			} catch (ExecutionException e) {
				log.error("Problem encountered during file transfer.", e);
				throw new FileSystemServiceException(e);
			}
		}
	}
}
